(function(global)
{
	function buildLogger(loggingFunction)
	{
		// Logging function that extends on what the built in Ignition one does
		// to print out the full contents of all arguments supplied.
		return function doLog(message)
		{
			var messages = Array.prototype.slice.call(arguments);
			var numArgs = messages.length;
			var i = 0;

			for (; i<numArgs; i++)
			{
				if (typeof messages[i] === 'object')
					messages[i] = JSON.stringify(messages[i], null, '\t');
			}

			loggingFunction(messages.join(' '));
		}
	}

	// Expose
	global.console =
	{
		log: buildLogger(log.console.debug),
		info: buildLogger(log.console.info),
		warn: buildLogger(log.console.warn),
		error: buildLogger(log.console.error)
	};
})(this);

(function(global)
{
	global.navigator =
	{
		// TODO: Determine proper values for these, and make them read only.
		appName: 'Ignition',
		appCodeName: 'Ignition',
		userAgent: 'Ignition User Agent',
		platform: 'Linux'
	};
})(this);
(function(global)
{
	var idCounter = 0;
	var timeoutsById = {};
	var intervalsById = {};

	function generateId()
	{
		return idCounter++;
	}

	function buildTimerInfo(id, func, delay, args)
	{
		var info =
		{
			id: id,
			func: func,
			delay: delay,
			args: Array.prototype.slice.call(args, 2),
			prevTime: Date.now()
		};

		return info;
	}

	function process(container, afterCall)
	{
		for (var id in container) if (container.hasOwnProperty(id))
		{
			var info = container[id];

			if (Date.now() - info.prevTime >= info.delay)
			{
				try
				{
					info.func.apply(global, info.args);
				}
				catch(e)
				{
					console.error("Exception leaked from setTimeout/setInterval callback:",
						e.message, ", stack:", e.stack);
				}
				afterCall(info);
			}
		}
	}

	function clearAllTimeouts()
	{
		for (var prop in timeoutsById)
		{
			if (timeoutsById.hasOwnProperty(prop))
			{
				delete timeoutsById[prop];
			}
		}

		for (var prop in intervalsById)
		{
			if (intervalsById.hasOwnProperty(prop))
			{
				delete intervalsById[prop];
			}
		}
	}

	function processAllTimeouts()
	{
		process(timeoutsById, deleteTimeout);
		process(intervalsById, updatePrevCallTime);
	}

	function bindTimeoutTick()
	{
		if (!timing.tick.has(processAllTimeouts, global))
		{
			timing.tick.add(
				processAllTimeouts,
				global,
				0,
				signals.SignalOwner.FRAMEWORK);
		}
	}

	function deleteTimeout(info)
	{
		delete timeoutsById[info.id];
	}

	function updatePrevCallTime(info)
	{
		info.prevTime = Date.now();
	}

	bindTimeoutTick();

	/**
	 * Creates a timeout to call the provided function after a delay.
	 *
	 * Note that the maximum resolution of the timer is bound by the frequency
	 * with which Ignition's timing.tick signal is being fired - if the maximum
	 * frequency of the signal is for it to fire every 50 ms then no timers will
	 * be fired more quickly than that.
	 *
	 * @see https://developer.mozilla.org/en-US/docs/Web/API/Window.setTimeout
	 * @param {Function} func
	 * @param {number} delay
	 * @returns {number}
	 */
	function setTimeout(func, delay)
	{
		var id = generateId();

		timeoutsById[id] = buildTimerInfo(id, func, delay, arguments);

		return id;
	}

	/**
	 * Clears the timeout associated with the provided ID.
	 *
	 * @see https://developer.mozilla.org/en-US/docs/Web/API/Window.clearTimeout
	 * @param {number} id
	 */
	function clearTimeout(id)
	{
		delete timeoutsById[id];
	}

	/**
	 * Creates an interval to call the provided function periodically.
	 *
	 * Note that the maximum resolution of the interval is bound by the frequency
	 * with which Ignition's timing.tick signal is being fired - if the maximum
	 * frequency of the signal is for it to fire every 50 ms then no intervals
	 * will be fired more quickly than that.
	 *
	 * @see https://developer.mozilla.org/en-US/docs/Web/API/Window.setInterval
	 * @param {Function} func
	 * @param {number} delay
	 * @returns {number}
	 */
	function setInterval(func, delay)
	{
		var id = generateId();

		intervalsById[id] = buildTimerInfo(id, func, delay, arguments);

		return id;
	}

	/**
	 * Clears the interval associated with the provided ID.
	 *
	 * @see https://developer.mozilla.org/en-US/docs/Web/API/Window.clearInterval
	 * @param {number} id
	 */
	function clearInterval(id)
	{
		delete intervalsById[id];
	}

	// Expose
	global.setTimeout = setTimeout;
	global.clearTimeout = clearTimeout;
	global.setInterval = setInterval;
	global.clearInterval = clearInterval;
	global.clearAllTimeouts = clearAllTimeouts;
})(this);
(function(global)
{
	const EVENTS =
	[
		'open',
		'close',
		'message',
		'fail'
	];

	function WebSocket(url)
	{
		this._client = new network.WebSocketClient();
		this._client.connect(url);

		EVENTS.forEach(_mapSignalToEvent.bind(this));
	}

	function _mapSignalToEvent(eventName)
	{
		var callbackName = 'on' + eventName;

		this._client[eventName].add(function()
		{
			if (typeof this[callbackName] === 'function')
				this[callbackName].apply(this, arguments);
		}.bind(this));
	}

	WebSocket.prototype.send = function(data)
	{
		this._client.send(data);
	};

	WebSocket.prototype.close = function(code)
	{
		this._client.disconnect(code);
	};

	// Expose
	global.WebSocket = WebSocket;
})(this);
(function(global)
{
	// Create a global object called window, which just points back to the
	// Ignition global object. This mimics the behaviour of the Window object
	// in the browser, which acts as global.
	global.window = global;

	// Expose window.atob() and window.btoa() though our plugin network
	if ( !!global['network'] )
	{
		if ( !!network['base64Encode'] && typeof(network.base64Encode) === 'function' )
			global.window.btoa = network.base64Encode;

		if ( !!network['base64Decode'] && typeof(network.base64Decode) === 'function' )
			global.window.atob = network.base64Decode;
	}

	global.location = global.window.location || {};

})(this);
