/* Copyright 2013-2018 Amazon.com, Inc. or its affiliates. All Rights Reserved. */

#ifndef PLUGIN_NETWORK_WEBSOCKETS_WEBSOCKET_DELEGATE_H_
#define PLUGIN_NETWORK_WEBSOCKETS_WEBSOCKET_DELEGATE_H_

#include <string>
#include <utils/UUID.h>

#ifndef PLUGINNETWORK_BUILD_LIBWEBSOCKET
// websocketpp
#define BOOST_ASIO_HAS_STD_CHRONO
#include "websocketpp/config/boost_config.hpp"
#include "websocketpp/config/asio_client.hpp"
#include "websocketpp/client.hpp"
#else /* PLUGINNETWORK_BUILD_LIBWEBSOCKET */
#include "DataStructures.h"
#include "core/thread/Mutex.h"

#include "libwebsockets.h"
#include "boost/regex.hpp"

#ifdef LIBWEBSOCKETS_CUSTOM_GETADDRINFO
void setup_lws_addrinfo_functions();
#endif //LIBWEBSOCKETS_CUSTOM_GETADDRINFO
#ifdef LIBWEBSOCKETS_CUSTOM_GETRANDOM
void setup_lws_random_function();
#endif //LIBWEBSOCKETS_CUSTOM_GETRANDOM

#endif /* PLUGINNETWORK_BUILD_LIBWEBSOCKET */

namespace boost
{
class thread;
}

namespace ignition
{
namespace network
{

enum WebSocketSendDataType
{
	WEBSOCKET_DATA_TYPE_TEXT = 0,
	WEBSOCKET_DATA_TYPE_BINARY
};

// This enum is just for libwebsockets.
// Websocketpp does not need this configuration, put it here for consistency in tests
enum WebSocketSSLConfiguration
{
	WEBSOCKET_SSL_CONFIGURATION_NO_SSL = 0,
	WEBSOCKET_SSL_CONFIGURATION_USE_SSL, // Same as LCCSCF_USE_SSL
	WEBSOCKET_SSL_CONFIGURATION_USE_SSL_ALLOW_SELFSIGNED // Same as LCCSCF_ALLOW_SELFSIGNED
};

enum WebSocketState
{
	WEBSOCKET_STATE_DISCONNECTED,
	WEBSOCKET_STATE_CONNECTED
};

class WebSocketDelegate
{
public:
	WebSocketDelegate(core::UUID uuid);
	virtual ~WebSocketDelegate();

	bool connect(const std::string& uri);
	void disconnect(unsigned int statusCode);
	bool send(const char* data, unsigned int length, WebSocketSendDataType type);
	const std::string& getUri();
	void setSSLFlag(int flag);

protected:

#if !defined(PLUGINNETWORK_BUILD_LIBWEBSOCKET) 
	typedef websocketpp::client<websocketpp::config::asio_tls_client> client;
	virtual void onOpen(websocketpp::connection_hdl hdl);
	virtual void onClose(websocketpp::connection_hdl hdl);
	virtual void onFail(websocketpp::connection_hdl hdl);
	virtual void onMessage(websocketpp::connection_hdl hdl, client::message_ptr msg);

	client _client;
	client::connection_ptr _connection;
#elif defined(PLUGINNETWORK_BUILD_LIBWEBSOCKET)
	lws* _websocket;
	lws_context* _context;
	lws_protocols _protocols[2];

	static int callback_entrance(lws* wsi, lws_callback_reasons reason, void* user, void* in, size_t len);

public:
	int callback_main(lws* wsi, lws_callback_reasons reason, void* user, void* in, size_t len);
protected:
	void Run();

	virtual void onOpen();
	virtual void onClose();
	virtual void onFail(void* msg, size_t len);
	virtual void onMessage(void* msg, size_t len);
#endif

	std::string _uri;
	core::UUID _uuid;
	WebSocketState _state;

	boost::thread* _clientThread;

	static unsigned int _instanceCounter;

private:
#if defined(PLUGINNETWORK_BUILD_LIBWEBSOCKET)
	static const boost::regex URI_REGEX;
	static const int BASE_TEN;
	static const int MAX_PORT_NUMBER;

	static const int SOCKET_WAIT_TIMEOUT = 100;
	static const int MAX_QUEUE_SIZE = 30;

	lws* _clientConnect();
	bool _validateURI(std::string& address, int& port, std::string& path);

	void _sendNextMessage();

	bool _shouldStopClientThread;
	bool _disconnectRequested;
	int _sslFlag;
	const std::string _protocolName;

	struct WebSocketData
	{
		std::string message;
		WebSocketSendDataType type;
	};

	core::queue<WebSocketData> _messageQueue;
	core::thread::Mutex _messageQueueMutex;
#endif
};

} // namespace network
} // namespace ignition

#endif // PLUGIN_NETWORK_WEBSOCKETS_WEBSOCKET_CLIENT_H_
