#version 100

precision highp float;

uniform sampler2D u_texture;
uniform vec4 u_color;

varying vec2 v_texcoord;
varying vec4 v_color;

const float kRefWhiteLevel = 100.0;

//ref: ARIB STD-B62 and BT.2087
const mat3 BT709_TO_BT2020 = mat3(
	0.6274,    0.3293,    0.0433,
	0.0691,    0.9195,    0.0114,
	0.0164,    0.0880,    0.8956
);

vec4 SRGB_EOTF(vec4 E)
{
	vec3 dark = E.xyz/12.92;
	vec3 light = pow((E.xyz+0.055)/(1.0+0.055), vec3(2.4, 2.4, 2.4));
	vec4 r;

	if (E.x <= 0.04045)
		r.x = dark.x;
	else
		r.x = light.x;

	if (E.y <= 0.04045)
		r.y = dark.y;
	else
		r.y = light.y;

	if (E.z <= 0.04045)
		r.z = dark.z;
	else
		r.z = light.z;

	r.w = E.w;
	return r;
}

vec4 SRGB_OETF(vec4 L)
{
	vec3 dark = L.xyz * 12.92;
    vec3 light = 1.055 * pow(L.xyz, vec3(1.0 / 2.4)) - 0.055;

    vec4 r;
	if (L.x <= 0.0031308)
		r.x = dark.x;
	else
		r.x = light.x;

	if (L.y <= 0.0031308)
		r.y = dark.y;
	else
		r.y = light.y;

	if (L.z <= 0.0031308)
		r.z = dark.z;
	else
		r.z = light.z;

	r.w = L.w;
    return r;
}

//apply gamma adjustment to (minL, maxL).
vec4 GammaAdjOOTF(vec4 L, float minLNits, float maxLNits, float gamma)
{
	vec3 nits = L.xyz * kRefWhiteLevel;
	vec4 i = vec4((nits - minLNits) / (maxLNits - minLNits), 1.0);
	vec3 j;
	j = pow( SRGB_OETF(i).xyz, vec3(gamma));

	vec3 adj =  (minLNits + (maxLNits - minLNits) * j) / kRefWhiteLevel;
	vec4 ret;
	if (nits.x >= minLNits && nits.x < maxLNits)
		ret.x = adj.x;
	else
		ret.x = L.x;

	if (nits.y >= minLNits && nits.y < maxLNits)
		ret.y = adj.y;
	else
		ret.y = L.y;

	if (nits.z >= minLNits && nits.z < maxLNits)
		ret.z = adj.z;
	else
		ret.z = L.z;

	ret.w = L.w;
	return ret;
}

//input: normalized L in units of RefWhite (1.0=100nits), output: normalized E
vec4 PQ_OETF(in vec4 L, float gamma)
{
	L = GammaAdjOOTF(L, 0.0, 300.0, gamma);

	const float c1 = 0.8359375;//3424.f/4096.f;
	const float c2 = 18.8515625;//2413.f/4096.f*32.f;
	const float c3 = 18.6875; //2392.f/4096.f*32.f;
	const float m1 = 0.159301758125; //2610.f / 4096.f / 4;
	const float m2 = 78.84375;// 2523.f / 4096.f * 128.f;
	L = L * kRefWhiteLevel / 10000.0;
	vec3 Lm1 = pow(L.xyz, vec3(m1,m1,m1));

	vec3 X = (c1 + c2 * Lm1) / (1.0 + c3 * Lm1);
	vec4 res = vec4(pow(X, vec3(m2,m2,m2)), L.w);
	return res;
}

vec4 Gamma(vec4 color)
{
	vec4 linearColor = SRGB_EOTF(color);

	// convert to BT2020
	vec4 c1 = vec4(linearColor.rgb*BT709_TO_BT2020,linearColor.a);
	vec4 clr = PQ_OETF(c1, 2.2);
	return vec4(clr.r, clr.g, clr.b, color.a);
}

void main()
{
	vec4 color = texture2D(u_texture, v_texcoord);
	vec4 result = color * v_color * u_color;
	gl_FragColor = Gamma(result);
}
